from __future__ import absolute_import

import doubleratchet

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

class CBCAEAD(doubleratchet.AEAD):
    CRYPTOGRAPHY_BACKEND = default_backend()

    def __init__(self):
        super(CBCAEAD, self).__init__()

    def __getHKDFOutput(self, message_key):
        # Prepare the salt, which should be a string of 0x00 bytes with the length of
        # the hash digest
        salt = b"\x00" * 32

        # Get 80 bytes from the HKDF calculation
        hkdf_out = HKDF(
            algorithm = hashes.SHA256(),
            length    = 80,
            salt      = salt,
            info      = "WhisperMessageKeys".encode("US-ASCII"),
            backend   = self.__class__.CRYPTOGRAPHY_BACKEND
        ).derive(message_key)

        # Split these 80 bytes in three parts
        return hkdf_out[:32], hkdf_out[32:64], hkdf_out[64:]

    def __getAES(self, key, iv):
        return Cipher(
            algorithms.AES(key),
            modes.CBC(iv),
            backend = self.__class__.CRYPTOGRAPHY_BACKEND
        )

    def encrypt(self, plaintext, message_key, ad):
        encryption_key, authentication_key, iv = self.__getHKDFOutput(message_key)

        # Prepare PKCS#7 padded plaintext
        padder    = padding.PKCS7(128).padder()
        plaintext = padder.update(plaintext) + padder.finalize()

        # Encrypt the plaintext using AES-256 (the 256 bit are implied by the key size),
        # CBC mode and the previously created key and iv
        aes_cbc    = self.__getAES(encryption_key, iv).encryptor()
        ciphertext = aes_cbc.update(plaintext) + aes_cbc.finalize()

        return {
            "ciphertext" : ciphertext,
            "additional" : {
                "key" : authentication_key,
                "ad"  : ad
            }
        }

    def decrypt(self, ciphertext, message_key, ad):
        decryption_key, authentication_key, iv = self.__getHKDFOutput(message_key)

        # Decrypt the plaintext using AES-256 (the 256 bit are implied by the key size),
        # CBC mode and the previously created key and iv
        aes_cbc   = self.__getAES(decryption_key, iv).decryptor()
        plaintext = aes_cbc.update(ciphertext) + aes_cbc.finalize()

        # Remove the PKCS#7 padding from the plaintext
        unpadder  = padding.PKCS7(128).unpadder()
        plaintext = unpadder.update(plaintext) + unpadder.finalize()

        return {
            "plaintext"  : plaintext,
            "additional" : {
                "key" : authentication_key,
                "ad"  : ad
            }
        }



from __future__ import absolute_import

import doubleratchet

class RootChain(doubleratchet.kdfchains.KDFChain):
    def __init__(self, key = None):
        super(RootChain, self).__init__(
            doubleratchet.recommended.RootKeyKDF(
                "SHA-256",
                "WhisperRatchet".encode("US-ASCII")
            ),
            key
        )



from __future__ import absolute_import

import doubleratchet

from .sendrecvchain import SendRecvChain

class SymmetricKeyRatchet(doubleratchet.ratchets.SymmetricKeyRatchet):
    def __init__(self):
        super(SymmetricKeyRatchet, self).__init__(
            SendRecvChain,
            SendRecvChain
        )



from __future__ import absolute_import

import omemo

class X3DHPKEncoder(omemo.backends.X3DHPKEncoder):
    @staticmethod
    def encodePublicKey(key, key_type):
        if key_type == "25519":
            return b"\x05" + key

        raise NotImplementedError

    @staticmethod
    def decodePublicKey(key_encoded):
        if key_encoded[0:1] == b"\x05":
            return (key_encoded[1:], "25519")
        
        raise NotImplementedError



from __future__ import absolute_import

import doubleratchet

class SendRecvChain(doubleratchet.kdfchains.ConstKDFChain):
    def __init__(self, key):
        super(SendRecvChain, self).__init__(
            None,
            doubleratchet.recommended.ChainKeyKDF("SHA-256"),
            key
        )



from __future__ import absolute_import

import doubleratchet
import x3dh

from .cbcaead import CBCAEAD
from .rootchain import RootChain
from .symmetrickeyratchet import SymmetricKeyRatchet

class DoubleRatchet(doubleratchet.ratchets.DoubleRatchet):
    def __init__(
        self,
        ad,
        root_key,
        own_key   = None,
        other_pub = None
    ):
        super(DoubleRatchet, self).__init__(
            CBCAEAD(),                              # aead
            5000,                                   # message_key_store_max
            SymmetricKeyRatchet(),                  # symmetric_key_ratchet
            ad,                                     # ad
            x3dh.implementations.KeyPairCurve25519, # encryption_key_pair_class
            RootChain(root_key),                    # root_chain
            own_key,
            other_pub
        )

    def _makeAD(self, header, ad):
        return ad

    def encryptMessage(self, *args, **kwargs):
        result = super(DoubleRatchet, self).encryptMessage(*args, **kwargs)
        result["additional"] = result["ciphertext"]["additional"]
        result["ciphertext"] = result["ciphertext"]["ciphertext"]
        return result



from setuptools import setup, find_packages

import os
import sys

version_file_path = os.path.join(
	os.path.dirname(os.path.abspath(__file__)),
	"omemo_backend_signal",
	"version.py"
)

version = {}

try:
	execfile(version_file_path, version)
except:
	with open(version_file_path) as fp:
		exec(fp.read(), version)

with open("README.md") as f:
    long_description = f.read()

setup(
    name = "omemo-backend-signal",
    version = version["__version__"],
    description = "A backend for python-omemo offering compatibility with libsignal.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/Syndace/python-omemo-backend-signal",
    author = "Tim Henkes",
    author_email = "tim@cifg.io",
    license = "GPLv3",
    packages = find_packages(),
    install_requires = [
        "X3DH>=0.5.6,<0.6",
        "DoubleRatchet>=0.6.0,<0.7",
        "OMEMO>=0.9.0,<0.10",
        "cryptography>=1.7.1",
        "protobuf>=2.6.1"
    ],
    python_requires = ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4",
    zip_safe = False,
    classifiers = [
        "Development Status :: 4 - Beta",

        "Intended Audience :: Developers",

        "Topic :: Communications :: Chat",
        "Topic :: Security :: Cryptography",

        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",

        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",

        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ]
)



from __future__ import absolute_import

from .x3dhpkencoder import X3DHPKEncoder

import x3dh

class State(x3dh.State):
    def __init__(self):
        """
        Sets the constructor parameters to the defaults used by OMEMO.
        The curve, min_num_otpks and max_num_otpks parameters were found in the XEP
        (https://xmpp.org/extensions/xep-0384.html).

        The hash_function and info_string parameters were found in the source code of
        libsignal-protocol-java
        (https://github.com/WhisperSystems/libsignal-protocol-java).

        The timeout for the SPK is defaulted to one week.
        """

        return super(State, self).__init__(
            "WhisperText".encode("US-ASCII"), # info_string
            "25519",                          # curve
            "SHA-256",                        # hash_function
            7 * 24 * 60 * 60,                 # spk_timeout
            20,                               # min_num_otpks
            100,                              # max_num_otpks
            X3DHPKEncoder                     # public_key_encoder_class
        )

    def getSharedSecretActive(self, *args, **kwargs):
        # X3DH is specified to build the associated data as follows: IK_A || IK_B.
        # As per usual, WhisperSystems weren't satisfied with their own solution and
        # instead of using the ad as built by X3DH they ALWAYS do:
        #     IK_sender || IK_receiver.
        # That means, when decrypting a message, a different ad is used as when encrypting
        # a message.
        #
        # To allow for this to happen, the ad is reordered so that it always has following
        # structure: IK_own || IK_other.

        result = super(State, self).getSharedSecretActive(*args, **kwargs)

        IK_own   = result["ad"][:33]
        IK_other = result["ad"][33:]

        result["ad"] = IK_own + IK_other

        return result

    def getSharedSecretPassive(self, *args, **kwargs):
        result = super(State, self).getSharedSecretPassive(*args, **kwargs)

        # See getSharedSecretActive for an explanation
        IK_own   = result["ad"][33:]
        IK_other = result["ad"][:33]

        result["ad"] = IK_own + IK_other

        return result
